(* renvoie la preuve de l'objectif dfini *)
type termes = const of string | var of string | func of string * (termes list);;
type type_preuve = but of termes | hyplst of type_preuve list | 
  buthypcpl of type_preuve * type_preuve;;
type substitution = sb of termes * termes;;
type unificateur = sl of substitution list | RIEN;;
type resultat = faux | vrai of unificateur * type_preuve;;

(* objectif test par ce programme *)
let objectif = func ("frere", [var "w";const "fiston"]);;

(* regles pour cet exercice *)
let regles = [
[func ("pere",[const "fiston";const "papa"])];
[func ("pere",[const "papa";const "papy"])];
[func ("pere",[const "fiston2";const "papa"])];
[func ("pere",[const "papa2";const "papy"])];
[func ("frere",[var "x";var "y"]) ; func ("pere",[var "x";var "z"]) ;
  func ("pere",[var "y";var "z"])];
[func ("oncle",[var "x";var "y"]) ; func ("pere",[var "x";var "z"]) ;
  func ("frere",[var "y";var "z"])];
[func ("gpere",[var "x";var "z"]) ;
  func ("pere",[var "x";var "y"]) ; func ("pere",[var "y";var "z"])]
];;

let rec affiche s = match s with
(* affiche le terme s *)
  const a -> print_string a
  | var x -> print_string x
  | func (f,l) -> print_string f;
                  print_string " (";
                  affparam l
and affparam l = match l with
(* affiche la liste des termes l *)
  [] -> print_string ")"
  | t::q -> affiche t;
            if q <> [] then print_char `,`;
            affparam q
;;

let rec print_preuve p = match p with
(* ecrit la type_preuve p sous la forme (h1 & (h3 -> h2) -> b) avec
  b est le but, h1 et h2 sont les hypotheses de ce but, -> h1 et h3 -> h2   *)
  but b -> affiche b                     (* affiche le but courant *)
  | buthypcpl (a,hyplst []) -> print_preuve a (* a est un fait *)
  | buthypcpl (a,b) -> print_char `(`;        (* affiche (b -> a) *)
                       print_preuve b;
                       print_string " -> ";
                       print_preuve a;
                       print_char `)`
  | hyplst (t::q) -> print_preuve t;
                     if q <> [] then (print_string " & ";
                     print_preuve (hyplst q))
  | _ -> failwith "cas inaccessible dans print_preuve"
;;

let print_regle r =
(* affiche une regle sous la forme hyp1 & hyp2 -> csq *)
  let rec print_regle_hyp r = match r with
    [] -> print_string "";
    | t::q -> affiche t;
              if q = [] then print_char ` ` else print_string " & ";
	      print_regle_hyp q         (* hypothese suivante *)
in match r with 
  [] -> failwith "Erreur : regle vide"
  | t::q -> print_char ` `;             (* indentation *)
            print_regle_hyp q;          (* affiche les hyp. de la regle *)
            print_string "-> ";
            affiche t                   (* affiche la conclusion *)
;;
let print_regles r =
(* affiche toutes les regles *)
  let rec print_regles_rec r = match r with
    [] -> print_string ""
    | t::q -> print_regle t;              (* affiche une regle *)
              print_char `\n`;            (* va a la ligne *)
              print_regles_rec q          (* affiche les regles suivantes *)
in print_string "Les rgles sont :\n";
   print_regles_rec r;
   print_char `\n`;
   flush std_out                        (* assure l'affichage effectif *)
;;

let rec applisubst s t = match t with
(* renvoie le terme t a qui on a appliqu la substitution s *)
  const a -> const a
  | var v -> (match s with
                sb (var a,b) -> if a = v then b else var v
                | _ -> var v
             )
  | func (f,l) -> (match s with
                     sb (func (ff,ll),b) -> if f = ff then b
                       else func (f,map (applisubst s) l)
                     | _ -> func (f,map (applisubst s) l)
                  )
;;

let rec appliunif u t = match u with
(* applique l'unificateur u au terme t *)
  sl [] -> t
  | RIEN -> t
  | sl (tt::qq) -> applisubst tt (appliunif (sl qq) t)
;;

let unifcpl u c = match c with
(* applique u au premier terme du couple c *)
  buthypcpl (but a,b) -> buthypcpl (but (appliunif u a),b)
  | _ -> failwith "mauvaise utilisation de unifcpl"
;;

let rec contient x b = match b with
(* renvoie true si le terme b contient la variable x, false sinon *)
  var y -> if var y = x then true else false
  | const _ -> false
  | func (f,l) -> contientliste x l
and contientliste x l = match l with
  [] -> false
  | t::q -> if contient x t then true else contientliste x q
;;

let rec unificateur t1 t2 = match (t1,t2) with
(* renvoie l'unificateur principal s'il existe, RIEN sinon *)
  (var a,var b) -> sl [sb (t1,t2)]
  | (var a,const b) -> sl [sb (t1,t2)]
  | (var a,func (f,l)) -> if contient (var a) t2 then RIEN else sl [sb (t1,t2)]
  | (const a,const b) -> if a = b then sl [] else RIEN
  | (func (f1,l1),func (f2,l2)) -> if f1 = f2 then unifieliste l1 l2
      else RIEN
  | (a,b) -> unificateur b a
and unifieliste l1 l2 = match (l1,l2) with
  ([],[]) -> sl []
  | ([],t2::q2) -> RIEN
  | (t1::q1,[]) -> RIEN
  | (t1::q1,t2::q2) -> let u = unifieliste q1 q2 in
    (match u with
      RIEN -> RIEN
      | sl l -> (match unificateur t1 t2 with
                  RIEN -> RIEN
                  | sl ll -> sl (ll@l)
                )
    )
;;

let unifiable t1 t2 =
(* renvoie true s'il existe un unificateur *)
  not (unificateur t1 t2 = RIEN)
;;

let rec prouve r rtot b = match r with
(* renvoie vrai(unificateur,preuvecpl(but,hyp)) si b est dmontrable, ou faux *)
  [] -> faux                    (* toutes les regles ont t testes *)
  | t::q -> match t with        (* essaye une regle *)
    [] -> failwith "Erreur : rgle vide"
    | x::q2 -> if unifiable x b then   (* si b est la csq de cette regles *)
      let u = unificateur x b in
(* essaye de prouver les hypotheses apres application de u *)
          (match prouvelist rtot (map (appliunif u) q2) with
            vrai (sl [],hyplst l) ->
(* hypotheses vrifies *)
vrai (u,buthypcpl(but (appliunif u b),hyplst l))
            | vrai (uu,_) -> failwith "prouvelist a renvoy un unificateur"
            | faux -> prouve q rtot b  (* essaye la regle suivante *)
          )
        else prouve q rtot b           (* idem *)
and prouvelist rtot l = match l with
(* renvoie vrai([],hypliste) si toutes les hypotheses sont prouvees *)
  [] -> vrai (sl [],hyplst [])
  | t::q -> (match prouve rtot rtot t with
              faux -> faux
              | vrai (u,cpl) ->
                (match prouvelist rtot (map (appliunif u) q) with
                  faux -> faux      (* propagation du faux *)
                  | vrai (_,hyplst hlq) ->
(* renvoie la preuve unifiee *)
vrai (sl [],hyplst ((unifcpl u cpl)::hlq))
                  | vrai (_,_) -> failwith "cas inaccessible"
                )
            )
;;

print_regles regles;
print_string "objectif : ";
affiche objectif;
print_char `\n`;
flush std_out;
(match prouve regles regles objectif with
  faux -> print_string "Le but est indmontrable\n"
  | vrai (s,p) -> print_string "Le but est prouv : \n";
                  print_preuve p;
		  print_char `\n`
)
;;
exit 0;;
