(* renvoie toutes les preuve de l'objectif dfini *)
type termes = const of string | var of string | func of string * (termes list);;
type substitution = sb of termes * termes;;
type unificateur_type = sl of substitution list | RIEN;;
type type_preuve = FAUX | but of termes | buthypcpl of type_preuve * type_preuve
  | res of (unificateur_type * type_preuve) list | hyplst of type_preuve list
  | lsthyplsteqv of type_preuve list;;

(* objectif test par ce programme *)
let objectif = func ("frere", [var "w";const "fiston"]);;

(* regles pour cet exercice *)
let regles = [
[func ("pere",[const "fiston";const "papa"])];
[func ("pere",[const "papa";const "papy"])];
[func ("pere",[const "fiston2";const "papa"])];
[func ("pere",[const "papa2";const "papy"])];
[func ("frere",[var "x";var "y"]) ; func ("pere",[var "x";var "z"]) ;
  func ("pere",[var "y";var "z"])];
[func ("oncle",[var "x";var "y"]) ; func ("pere",[var "x";var "z"]) ;
  func ("frere",[var "y";var "z"])];
[func ("gpere",[var "x";var "z"]) ;
  func ("pere",[var "x";var "y"]) ; func ("pere",[var "y";var "z"])]
];;

let rec affiche s = match s with
(* affiche le terme s *)
  const a -> print_string a
  | var x -> print_string x
  | func (f,l) -> print_string f;
                  print_string " (";
                  affparam l
and affparam l = match l with
(* affiche la liste des termes l *)
  [] -> print_string ")"
  | t::q -> affiche t;
            if q <> [] then print_char `,`;
            affparam q
;;

let rec print_preuve p = match p with
(* ecrit la type_preuve p sous la forme (h1 & (h3 -> h2) -> b) avec
   b est le but, h1 et h2 sont les hypotheses de ce but, -> h1 et h3 -> h2    *)
  but b -> affiche b                     (* affiche le but courant            *)
  | buthypcpl (a,hyplst []) -> print_preuve a (* a est un fait                *)
  | buthypcpl (a,b) -> print_char `(`;        (* affiche (b -> a)             *)
                       print_preuve b;
                       print_string " -> ";
                       print_preuve a;
                       print_char `)`
  | hyplst (t::q) -> print_preuve t;
                     if q <> [] then (print_string " & ";
                     print_preuve (hyplst q))
  | res (_,t::q) -> print_preuve t;
                    if q <> [] then (print_string " ou ";
                                     print_preuve (res (sl [],q))
                                    )
  | _ -> failwith "cas inaccessible dans print_preuve"
;;

let print_regle r =
(* affiche une regle sous la forme hyp1 & hyp2 -> csq                         *)
  let rec print_regle_hyp r = match r with
    [] -> print_string "";
    | t::q -> affiche t;
              if q = [] then print_char ` ` else print_string " & ";
              print_regle_hyp q         (* hypothese suivante                 *)
in match r with 
  [] -> failwith "Erreur : regle vide"
  | t::q -> print_char ` `;             (* indentation                        *)
            print_regle_hyp q;          (* affiche les hyp. de la regle       *)
            print_string "-> ";
            affiche t                   (* affiche la conclusion              *)
;;
let print_regles r =
(* affiche toutes les regles                                                  *)
  let rec print_regles_rec r = match r with
    [] -> print_string ""               (* plus de regles                     *)
    | t::q -> print_regle t;            (* affiche une regle                  *)
              print_char `\n`;          (* va a la ligne                      *)
              print_regles_rec q        (* affiche le reste des regles        *)
in print_string "Les rgles sont :\n";
   print_regles_rec r;
   print_char `\n`;                     (* saute une ligne                    *)
   flush std_out                        (* assure l'affichage effectif        *)
;;
(*
let unifcpl u c = match c with
(* applique u au premier terme du couple c                                    *)
  buthypcpl (but a,b) -> buthypcpl (but (appliunif u a),b)
  | _ -> failwith "mauvaise utilisation de unifcpl"
;;
*)
let rec contient x b = match b with
(* renvoie true si le terme b contient la variable x, false sinon             *)
  var y -> if var y = x then true else false
  | const _ -> false
  | func (f,l) -> contientliste x l
and contientliste x l = match l with
  [] -> false
  | t::q -> if contient x t then true else contientliste x q
;;

let rec unificateur t1 t2 = match (t1,t2) with
(* renvoie l'unificateur principal s'il existe, RIEN sinon                    *)
  (var a,var b) -> sl [sb (t1,t2)]
  | (var a,const b) -> sl [sb (t1,t2)]
  | (var a,func (f,l)) -> if contient (var a) t2 then RIEN else sl [sb (t1,t2)]
  | (const a,const b) -> if a = b then sl [] else RIEN
  | (func (f1,l1),func (f2,l2)) -> if f1 = f2 then unificateurliste l1 l2
                                              else RIEN
  | (const a,func (f,l)) -> RIEN
  | _ -> unificateur t2 t1              (* essaye dans l'autre sens           *)
and unificateurliste l1 l2 = match (l1,l2) with
  ([],[]) -> sl []
  | ([],t2::q2) -> RIEN
  | (t1::q1,[]) -> RIEN
  | (t1::q1,t2::q2) -> (match unificateurliste q1 q2 with
    RIEN -> RIEN
    | sl l -> (match unificateur t1 t2 with
      RIEN -> RIEN
      | sl ll -> sl (ll@l)
    )
  )
;;

let unifiable t1 t2 =
(* renvoie true s'il existe un unificateur                                    *)
  not (unificateur t1 t2 = RIEN)
;;

let rec applisubst s t =
(* renvoie le terme t aprs application de la substitution s                  *)
  if fst (s) = t then snd (s) else
  match t with
    func (f,l) -> func (f,map (applisubst s) l)
    | _ -> t
;;

let rec appliunif u t = match u with
(* applique l'unificateur u au terme t, u tant une liste de substitutions    *)
  sl [] -> t
  | RIEN -> t                           (* cas impossible                     *)
  | sl (tt::qq) -> applisubst tt (appliunif (sl qq) t)
;;

let rec appliunifpreuve u p = match p with
(* applique rcursivement l'unificateur u a la preuve p                       *)
  but t -> appliunif u p
  | hyplst [] -> hyplst []
  | hyplst (t::q) -> (appliunifpreuve u t)::(appliunifpreuve u q)
  | buthypcpl (a,b) -> buthypcpl(appliunifpreuve a, appliunifpreuve b)
  | res [] -> res []
  | res ((a,b)::q) -> res ((a,appliunifpreuve u b)::(appliunifpreuve u (res q)))
  | _ -> p
;;

let rec prouvebutexhaust r rtot b = match r with
(* renvoie res [(unif,preuvecpl (but,hyp))] si b est prouv, FAUX sinon       *)
  [] -> FAUX                            (* toutes les regles ont t testes  *)
  | t::q -> match t with                (* essaye une regle                   *)
    [] -> failwith "rgle vide"         (* cas impossible                     *)
    | x::q2 -> if unifiable x b then    (* si b est la csq de cette regles    *)
      let u = unificateur x b in        (* applique u aux hypotheses          *)
        (match prouvelist rtot (map (appliunif u) q2) with
	  FAUX -> prouvebutexhaust q rtot b  (* regle fausse : teste la suite *)
          | lsthyplsteqv he -> construitlsthyplsteqv u b he;
            (* construit lhle [res(uu=u::u(he),cpl(but,h(he)))]    *)
                                        (* pc est la preuve courante          *)
            (match prouvebutexhaust q rtot b with (* a-t-on d'autres preuves? *)
              FAUX -> res [(u,pc)]      (* non : on renvoie [pc]              *)
              | res l -> res ((u,pc)::l)(* oui : rajoute pc                   *)
              | _ -> failwith "cas inaccessible"
            )
          )
               else prouvebutexhaust q rtot b        (* regle suivante        *)
and prouvelist rtot l = match l with
(* renvoie lsthyplsteqv[res (u,hyplst)], FAUX sinon      *)
  [] -> []
  | hyplst (t::q) -> (match prouvebutexhaust rtot rtot t with
    FAUX -> FAUX
    | res r -> (match prouvelist rtot q with
      FAUX -> FAUX
      | [] -> pltype []
      | pltype pl -> pltype ((prouvehypres (hyplist q) r)::pl)
    )
  )
and prouvelistres l r = match r with
  res [] -> res []
  | res ((u::bhc)::q) -> appliunifpreuve u l
;;

print_regles regles;
print_string "objectif : ";
affiche objectif;
print_char `\n`;
flush std_out;
(match prouvebutexhaust regles regles objectif with
  faux -> print_string "Le but est indmontrable\n"
  | vrai (s,p) -> print_string "Le but est prouv : \n";
                  print_preuve p;
		  print_char `\n`
)
;;
exit 0;;
